require 'sinatra'
require 'sinatra/namespace'
require 'json'
require 'rox/server/flags/rox_flag'
require 'rox/server/rox_server'
require 'rox/server/rox_options'

class Container
  attr_accessor :first_flag
  attr_reader :bool_default_false, :bool_default_true

  def initialize
    @bool_default_false = Rox::Server::RoxFlag.new(false)
    @bool_default_true = Rox::Server::RoxFlag.new(true)
    @first_flag = Rox::Server::RoxFlag.new(true)
  end
end

class ServerLogger
  def initialize
    @logger = Logger.new($stdout)
  end

  def debug(message, exception = nil)
    if exception.nil?
      @logger.debug(message)
    else
      @logger.debug("#{message}. Exception: #{exception}")
    end
  end

  def error(message, exception = nil)
    if exception.nil?
      @logger.error(message)
    else
      @logger.error("#{message}. Exception: #{exception}")
    end
  end

  def warn(message, exception = nil)
    if exception.nil?
      @logger.warn(message)
    else
      @logger.warn("#{message}. Exception: #{exception}")
    end
  end
end

container = Container.new
set :port, ARGV[0]

namespace '/' do
  before do
    content_type 'application/json'
  end

  get 'status-check' do
    [200, '']
  end

  get 'api/values/:id' do |id|
    if container.first_flag.enabled?
      [200, JSON.generate({ value: id })]
    else
      [200, 'Eladddddd']
    end
  end

  put 'api/values/:id' do
    [200, '']
  end

  delete 'api/values/:id' do
    [200, '']
  end

  post '' do
    data = JSON.parse(request.body.read)
    action = data['action']
    payload = data['payload']
    puts data

    case action
    when 'staticFlagIsEnabled'
      flag = payload['flag']
      result = container.send(flag.to_sym).enabled?(payload['context'])
      [200, JSON.generate({ result: result })]
    when 'registerStaticContainers'
      Rox::Server::RoxServer.register(container)
      [200, JSON.generate({ result: 'done' })]
    when 'setCustomPropertyToThrow'
      def raise_(ex)
        raise ex
      end
      Rox::Server::RoxServer.set_custom_string_property(payload['key'], raise_(Exception('error')))
      [200, JSON.generate({ result: 'done' })]
    when 'setCustomStringProperty'
      Rox::Server::RoxServer.set_custom_string_property(payload['key'], payload['value'])
      [200, JSON.generate({ result: 'done' })]
    when 'dynamicFlagValue'
      if payload['context']
        puts "There is a payload[context] => #{payload['context']}"
        payload['context'].each do |context|
          key = context[0]
          value = context[1]
          Rox::Server::RoxServer.set_custom_string_property(key, value) if value.instance_of? String
          if value.instance_of? TrueClass
            Rox::Server::RoxServer.set_custom_boolean_property(key, value)
          elsif value.instance_of? FalseClass
            Rox::Server::RoxServer.set_custom_boolean_property(key, value)
          end
          Rox::Server::RoxServer.set_custom_int_property(key, value) if value.instance_of? Integer
          Rox::Server::RoxServer.set_custom_float_property(key, value) if value.instance_of? Float
        end
      end
      result = Rox::Server::RoxServer.dynamic_api.value(payload['flag'], payload['defaultValue'], payload['context'],
                                                        [])
      [200, JSON.generate({ result: result })]
    when 'dynamicFlagIsEnabled'
      payload['context']&.each do |context|
        key = context[0]
        value = context[1]
        Rox::Server::RoxServer.set_custom_string_property(key, value) if value.instance_of? String
        if value.instance_of? TrueClass
          Rox::Server::RoxServer.set_custom_boolean_property(key, value)
        elsif value.instance_of? FalseClass
          Rox::Server::RoxServer.set_custom_boolean_property(key, value)
        end
        Rox::Server::RoxServer.set_custom_int_property(key, value) if value.instance_of? Integer
        Rox::Server::RoxServer.set_custom_float_property(key, value) if value.instance_of? Float
      end
      result = Rox::Server::RoxServer.dynamic_api.enabled?(payload['flag'], payload['defaultValue'], payload['context'])
      [200, JSON.generate({ result: result })]
    when 'setupAndAwait'
      env = 'stam'
      if payload['options']
        options = payload['options']
        if options['configuration']
          configuration = options['configuration']
          env = configuration['env']
        end
      end
      case env
      when 'qa'
        ENV['ROLLOUT_MODE'] = 'QA'
      when 'localhost'
        ENV['ROLLOUT_MODE'] = 'LOCAL'
      end
      options = Rox::Server::RoxOptions.new(logger: ServerLogger.new)
      puts "options => #{options}"
      Rox::Server::RoxServer.setup(payload['key'], options).value
      [200, JSON.generate({ result: 'done' })]
    when 'stop'
      Thread.new { sleep 1; Process.kill 'INT', Process.pid }
      halt 200, JSON.generate({ result: 'done' })
    end
  end
end
