require 'minitest/autorun'
require 'rox/server/rox_options'
require 'rox/server/rox_server'
require_relative 'container'
require_relative 'custom_props'
require_relative 'test_vars'

module E2E
  class Logger
    def debug(message, _ex = nil)
      puts 'Before Rox.Setup', message
    end

    def error(message, ex = nil)
      puts 'Before Rox.Setup', message, ex
    end

    def warn(message, _ex = nil)
      puts 'Before Rox.Setup', message
    end
  end

  class RoxE2ETest < Minitest::Test
    ENV['ROLLOUT_MODE'] = 'QA'

    configuration_fetched_handler = proc do |e|
      if !e.nil? && e.fetcher_status == Rox::Core::FetcherStatus::APPLIED_FROM_NETWORK
        TestVars.configuration_fetched_count += 1
      end
    end

    impression_handler = proc do |e|
      if !e.nil? && !e.reporting_value.nil? && (e.reporting_value.name == 'flag_for_impression')
        TestVars.is_impression_raised = true
      end
      TestVars.impression_returned_args = e
    end

    option = Rox::Server::RoxOptions.new(
      configuration_fetched_handler: configuration_fetched_handler,
      impression_handler: impression_handler,
      dev_mode_key: '67e39e708444aa953414e444',
      logger: Logger.new
    )

    @@container = Container.new
    Rox::Server::RoxServer.register(@@container)
    CustomProps.create_custom_props
    Rox::Server::RoxServer.setup('5b82864ebc3aec37aff1fdd5', option).join

    def test_simple_flag
      assert_equal true, @@container.simple_flag.enabled?
    end

    def test_simple_flag_overwritten
      assert_equal false, @@container.simple_flag_overwritten.enabled?
    end

    def test_string
      assert_equal 'red', @@container.variant.value
    end

    def test_variant_overwritten
      assert_equal 'green', @@container.variant_overwritten.value
    end

    def test_all_custom_properties
      assert_equal true, @@container.flag_custom_properties.enabled?

      assert_equal true, TestVars.is_computed_boolean_prop_called
      assert_equal true, TestVars.is_computed_float_prop_called
      assert_equal true, TestVars.is_computed_int_prop_called
      assert_equal true, TestVars.is_computed_semver_prop_called
      assert_equal true, TestVars.is_computed_string_prop_called
    end

    def test_fetch_within_timeout
      number_of_config_fetches = TestVars.configuration_fetched_count
      timeout = 5
      thread = Rox::Server::RoxServer.fetch
      res = thread.join(timeout)

      if res.nil?
        flunk('timeout')
      else
        assert number_of_config_fetches < TestVars.configuration_fetched_count
      end
    end

    def test_string_with_context
      some_positive_context = { 'isDuckAndCover' => true }
      some_negative_context = { 'isDuckAndCover' => false }

      assert_equal 'red', @@container.variant_with_context.value

      assert_equal 'blue', @@container.variant_with_context.value(some_positive_context)
      assert_equal 'red', @@container.variant_with_context.value(some_negative_context)
    end

    def test_target_groups_all_any_none
      TestVars.target_group1 = true
      TestVars.target_group2 = true

      assert_equal true, @@container.flag_target_groups_all.enabled?
      assert_equal true, @@container.flag_target_groups_any.enabled?
      assert_equal false, @@container.flag_target_groups_none.enabled?

      TestVars.target_group1 = false
      assert_equal false, @@container.flag_target_groups_all.enabled?
      assert_equal true, @@container.flag_target_groups_any.enabled?
      assert_equal false, @@container.flag_target_groups_none.enabled?

      TestVars.target_group2 = false
      assert_equal false, @@container.flag_target_groups_all.enabled?
      assert_equal false, @@container.flag_target_groups_any.enabled?
      assert_equal true, @@container.flag_target_groups_none.enabled?
    end

    def test_impression_handler
      @@container.flag_for_impression.enabled?
      assert_equal true, TestVars.is_impression_raised
      TestVars.is_impression_raised = false

      context = { 'var' => 'val' }
      flag_impression_value = @@container.flag_for_impression_with_experiment_and_context.enabled?(context)
      refute_nil TestVars.impression_returned_args
      refute_nil TestVars.impression_returned_args.reporting_value
      assert_equal true, TestVars.impression_returned_args.reporting_value.value
      assert_equal true, flag_impression_value
      assert_equal 'flag_for_impression_with_experiment_and_context',
                   TestVars.impression_returned_args.reporting_value.name

      refute_nil TestVars.impression_returned_args

      assert_equal 'val', TestVars.impression_returned_args.context['var']
    end

    def test_flag_dependency
      TestVars.is_prop_for_target_group_for_dependency = true
      assert_equal true, @@container.flag_for_dependency.enabled?
      assert_equal false, @@container.flag_dependent.enabled?

      TestVars.is_prop_for_target_group_for_dependency = false
      assert_equal true, @@container.flag_dependent.enabled?
      assert_equal false, @@container.flag_for_dependency.enabled?
    end

    def test_string_dependency_with_context
      some_positive_context = { 'isDuckAndCover' => true }
      some_negative_context = { 'isDuckAndCover' => false }

      assert_equal 'White', @@container.flag_color_dependent_with_context.value
      assert_equal 'White', @@container.flag_color_dependent_with_context.value(some_negative_context)
      assert_equal 'Yellow', @@container.flag_color_dependent_with_context.value(some_positive_context)
    end
  end
end
