require 'digest'
require 'json'

require 'rox/core/consts/environment'
require 'rox/core/consts/property_type'
require 'rox/core/logging/logging'
require 'rox/core/network/request'
require 'rox/core/network/request_data'
require 'rox/core/network/response'
require 'rox/core/utils/debouncer'

module Rox
  module Core
    class StateSender
      def initialize(sdk_settings, device_properties, flag_repository, custom_property_repository)
        @sdk_settings = sdk_settings
        @device_properties = device_properties
        @flag_repository = flag_repository
        @custom_property_repository = custom_property_repository
        @request = Request.new
        @debouncer = Debouncer.new(1, proc { send })
      end

      def delayed_send
        @debouncer.call
      end

      def send
        rollout_key = @device_properties.rollout_key
        serialized_feature_flags = StateSender.seralize_flag_repository(@flag_repository)
        serialized_custom_properties = StateSender.serialize_custom_properties_repository(@custom_property_repository)

        state_payload = StateSender.state_payload(serialized_feature_flags, serialized_custom_properties,
                                                  @device_properties, @sdk_settings.dev_mode_secret)
        md5_signature = StateSender.md5_signature(serialized_feature_flags, serialized_custom_properties,
                                                  @device_properties, @sdk_settings.dev_mode_secret)

        unless @device_properties.rox_options.self_managed?
          response = get_state_from_CDN(rollout_key, md5_signature)
          if response.success?
            Rox::Core::Logging.logger.debug('Successfully fetched state from CDN')
            return
          else
            Rox::Core::Logging.logger.debug('Failed to fetch state from CDN. Sending state to API...')
          end
        end

        response = send_state_to_API(rollout_key, md5_signature, state_payload)
        message = if response.success?
                    'Successfully sent state to API.'
                  else
                    'Failed to send state to API.'
                  end
        Rox::Core::Logging.logger.debug(message)
      end

      def dump_state
        serialized_feature_flags = StateSender.seralize_flag_repository_with_values(@flag_repository)
        serialized_custom_properties = StateSender.serialize_custom_properties_repository_with_values(@custom_property_repository)
        StateSender.state_payload(serialized_feature_flags, serialized_custom_properties, @device_properties,
                                  @sdk_settings.dev_mode_secret)
      end

      def get_state_from_CDN(rollout_key, md5_signature)
        url_for_state_cdn =
          RequestData.new("#{Rox::Core::Environment.state_cdn_path}/#{rollout_key}/#{md5_signature}", {})
        @request.send_get(url_for_state_cdn)
      end

      def send_state_to_API(rollout_key, md5_signature, state_payload)
        @request.send_post(
          "#{Rox::Core::Environment.state_api_path(@device_properties.rox_options.self_managed_options&.server_url)}/#{rollout_key}/#{md5_signature}", state_payload
        )
      end

      def self.md5_signature(serialized_feature_flags, serialized_custom_properties, device_properties, dev_mode_secret)
        values = state_payload(serialized_feature_flags, serialized_custom_properties, device_properties,
                               dev_mode_secret).values

        hash = Digest::MD5.hexdigest(values.join('|'))
        hash.upcase
      end

      def self.state_payload(seralized_flag_repository, serialized_custom_property_repository, device_properties, dev_mode_secret)
        values = {}
        values[PropertyType::APP_KEY.name] = device_properties.all_properties[PropertyType::APP_KEY.name]
        values[PropertyType::PLATFORM.name] = device_properties.all_properties[PropertyType::PLATFORM.name]
        values.merge!(device_properties.all_properties)
        values.merge!({ feature_flags: seralized_flag_repository })
        values.merge!({ custom_properties: serialized_custom_property_repository })
        values['devModeSecret'] = dev_mode_secret
        values
      end

      def self.seralize_flag_repository(flag_repository)
        flags = []
        flag_repository.all_flags.sort_by(&:name).each do |f|
          flags << {
            name: f.name,
            defaultValue: f.default_value,
            options: f.options
          }
        end
        flags
      end

      def self.seralize_flag_repository_with_values(flag_repository)
        flags = []
        flag_repository.all_flags.sort_by(&:name).each do |f|
          flags << {
            name: f.name,
            enabled: f.enabled?,
            defaultValue: f.default_value,
            options: f.options
          }
        end
        flags
      end

      def self.serialize_custom_properties_repository(custom_property_repository)
        properties = []
        custom_property_repository.all_custom_properties.sort_by(&:name).each do |p|
          properties << {
            name: p.name,
            type: p.type.type,
            externalType: p.type.external_type
          }
        end
        properties
      end

      def self.serialize_custom_properties_repository_with_values(custom_property_repository)
        properties = []
        custom_property_repository.all_custom_properties.sort_by(&:name).each do |p|
          properties << {
            name: p.name,
            value: p.value(nil),
            type: p.type.type,
            externalType: p.type.external_type
          }
        end
        properties
      end
    end
  end
end
