require 'securerandom'
require 'rox/core/core'
require 'rox/core/logging/logging'
require 'rox/server/rox_options'
require 'rox/server/client/sdk_settings'
require 'rox/server/client/server_properties'
require 'rox/server/flags/server_entities_provider'
require 'rox/core/properties/property_factory'

module Rox
  module Server
    class RoxServer
      @core = Rox::Core::Core.new

      class << self
        def setup(api_key, rox_options = nil)
          validate_api_key(api_key)

          rox_options = Rox::Server::RoxOptions.new if rox_options.nil?
          sdk_settings = Rox::Server::SdkSettings.new(api_key, rox_options.dev_mode_key)
          server_properties = Rox::Server::ServerProperties.new(sdk_settings, rox_options)

          RoxServer.add_custom_properties(Rox::Core::PropertyFactory.new(server_properties))

          @@thread = Thread.new do
            Thread.current.report_on_exception = false if Thread.current.respond_to?(:report_on_exception)
            @core.setup(sdk_settings, server_properties).join
          end
        end

        def shutdown
          return if !defined?(@@thread) || @@thread.nil?

          Thread.kill(@@thread)
          @core.shutdown
        end

        def add_custom_properties(property_factory)
          property_factory.all_properties.each do |property|
            @core.add_custom_property_if_not_exists(property)
          end
        end

        def register(rox_container)
          @core.register(rox_container)
        end

        def use_userspace_unhandled_error_handler(&handler)
          @core.userspace_unhandled_error_handler = handler
        end

        def context=(context)
          @core.context = context
        end

        def fetch
          Thread.new do
            Thread.current.report_on_exception = false if Thread.current.respond_to?(:report_on_exception)
            begin
              @core.fetch.join
            rescue StandardError => e
              Rox::Core::Logging.logger.error('Failed in Rox.fetch', e)
            end
          end
        end

        def set_custom_string_property(name, value = nil, &block)
          @core.add_custom_property(Rox::Core::CustomProperty.new(name, Rox::Core::CustomPropertyType::STRING, value,
                                                                  &block))
        end

        def set_custom_boolean_property(name, value = nil, &block)
          @core.add_custom_property(Rox::Core::CustomProperty.new(name, Rox::Core::CustomPropertyType::BOOL, value,
                                                                  &block))
        end

        def set_custom_int_property(name, value = nil, &block)
          @core.add_custom_property(Rox::Core::CustomProperty.new(name, Rox::Core::CustomPropertyType::INT, value,
                                                                  &block))
        end

        def set_custom_float_property(name, value = nil, &block)
          @core.add_custom_property(Rox::Core::CustomProperty.new(name, Rox::Core::CustomPropertyType::FLOAT, value,
                                                                  &block))
        end

        def set_custom_semver_property(name, value = nil, &block)
          @core.add_custom_property(Rox::Core::CustomProperty.new(name, Rox::Core::CustomPropertyType::SEMVER, value,
                                                                  &block))
        end

        def dynamic_api
          @core.dynamic_api(Rox::Server::ServerEntitiesProvider.new)
        end

        def dump_state
          @core.dump_state
        end

        private

        def validate_api_key(api_key)
          raise ArgumentError, 'api_key is not valid' if (api_key =~ /^[a-f\d]{24}$/).nil?

          true
        end
      end
    end
  end
end
